@@ -4,10 +4,11 @@ import datetime
 from uuid import UUID, uuid1
 from typing import Dict, Any, Optional, Tuple, Callable
 from importlib.metadata import version
-from urllib3.exceptions import InsecureRequestWarning
-
-import requests
-from requests.exceptions import RequestException
+import urllib.request
+import urllib.response
+import urllib.parse
+from urllib.error import URLError, HTTPError
+import ssl
 
 from rkclient.entities import PEM, Artifact
 from rkclient.serialization import PEMSerialization, _encode_as_base64
@@ -17,11 +18,20 @@ RK_VERSION = version('RecordKeeper_Client')
 
 
 def add_pem_context(files: Dict[str, Any], pem: PEM):
+    """
+    Helper function to be used when adding RK context to request
+    :param files: a dictionary which belongs to Python Requests object
+    :return:
+    """
     files['rk_context'] = pem.ID.hex
 
 
-# request == Flask request
 def get_pem_from_request(request) -> Optional[UUID]:
+    """
+    Helper function to be used when retrieving RK context
+    :param request: Flask-request object
+    :return:
+    """
     rk_context = request.files.get('rk_context', None)
     if rk_context is None:
         return None
@@ -35,20 +45,27 @@ class RKClient:
         If bool is False, str contains error description
 
         Errors are also logged to rkclient logger
-
     """
 
-    def __init__(self, receiver_url: str, emitter_id: Optional[UUID] = None, timeout_sec: int = 5, insecure: bool = True):
+    def __init__(self, receiver_url: str,
+                 emitter_id: Optional[UUID] = None,
+                 timeout_sec: int = 5,
+                 insecure: bool = True):
+        """
+
+        :param receiver_url:
+        :param emitter_id:
+        :param timeout_sec:
+        :param insecure: set it to True when operating with server that has test SSL certificates
+        """
         self.receiver_url = receiver_url.rstrip('/')
         if emitter_id is None:
             emitter_id = uuid1()
         self.emitter_id = emitter_id
         self.timeout_sec = timeout_sec
-        self.session = requests.Session()
         log.info(f"ver {RK_VERSION}, connecting to: {self.receiver_url}")
-        if insecure:
-            self.session.verify = False
-            requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)
+        self.insecure = insecure
+        if self.insecure:
             log.warning("Disabled SSL certificate check")
 
     def __enter__(self):
@@ -112,20 +129,21 @@ class RKClient:
         Sends PEM to Record Keeper.
         :return: check class description
         """
-        payload: str = PEMSerialization.to_json(pem)
-
         def _send_pem():
-            return self.session.post(url=self.receiver_url + "/pem", data=payload, timeout=self.timeout_sec)
-        log.debug(f"sending PEM: {payload}")
-        return _handle_request(_send_pem, "sending PEM")
+            payload: str = PEMSerialization.to_json(pem)
+            log.debug(f"sending PEM: {payload}")
+            return self._post("/pem", payload)
+
+        return _handle_request(_send_pem, "Sending PEM")
 
     def ping(self) -> Tuple[str, bool]:
         """
         :return: check class description
         """
         def _ping():
-            return self.session.get(url=self.receiver_url + "/ping", timeout=self.timeout_sec)
-        return _handle_request(_ping, "pinging")
+            return self._get("/ping")
+
+        return _handle_request(_ping, "Pinging")
 
     def get_info(self) -> Tuple[str, bool]:
         """
@@ -134,19 +152,20 @@ class RKClient:
         TODO: return object instead of str
         """
         def _get_info():
-            return self.session.get(url=self.receiver_url + "/info", timeout=self.timeout_sec)
-        return _handle_request(_get_info, "getting info", True)
+            return self._get("/info", )
+
+        return _handle_request(_get_info, "Getting info", True)
 
     def get_tag(self, namespace: UUID, tag_name: str) -> Tuple[str, bool]:
         """
         Returned tag is UUID in hex (do: UUID(hex=result))
         :return: check class description
         """
-        tag_base64 = _encode_as_base64(tag_name)
-
         def _get_tag():
-            return self.session.get(url=self.receiver_url + f"/tag/{namespace.hex}/{tag_base64}", timeout=self.timeout_sec)
-        return _handle_request(_get_tag, "getting tag", True)
+            tag_base64 = _encode_as_base64(tag_name)
+            return self._get(f"/tag/{namespace.hex}/{tag_base64}")
+
+        return _handle_request(_get_tag, "Getting tag", True)
 
     def set_tag(self, namespace: UUID, tag_name: str, pem: PEM) -> Tuple[str, bool]:
         """
@@ -156,29 +175,57 @@ class RKClient:
         :param pem:
         :return: check class description
         """
-        tag_base64 = _encode_as_base64(tag_name)
-
         def _set_tag():
-            return self.session.post(url=self.receiver_url + f"/tag/{namespace.hex}/{tag_base64}/{pem.ID.hex}", timeout=self.timeout_sec)
-        return _handle_request(_set_tag, "setting tag")
+            tag_base64 = _encode_as_base64(tag_name)
+            return self._post(f"/tag/{namespace.hex}/{tag_base64}/{pem.ID.hex}", tag_base64)
+
+        return _handle_request(_set_tag, "Setting tag")
+
+    def _get(self, url_postfix: str, query_params: Optional[Dict[str, Any]] = None) -> Tuple[str, bool]:
+        url = self.receiver_url + url_postfix + self._encode_query_params(query_params)
+        req = urllib.request.Request(url=url)
+        return self._make_request(req)
+
+    def _post(self, url_postfix: str, payload: str, query_params: Optional[Dict[str, Any]] = None) -> Tuple[str, bool]:
+        url = self.receiver_url + url_postfix + self._encode_query_params(query_params)
+        req = urllib.request.Request(url=url, data=payload.encode())
+        req.add_header('Content-Type', 'application/json')
+        return self._make_request(req)
+
+    def _make_request(self, request):
+        try:
+            resp = urllib.request.urlopen(request, timeout=self.timeout_sec, context=self._get_ssl_context())
+        except HTTPError as e:
+            return f"error: {e}", False
+        except URLError as e:
+            return f"connection error: {e.reason}", False
+        else:
+            return resp.read().decode(), True
+
+    def _get_ssl_context(self) -> ssl.SSLContext:
+        ctx = ssl.create_default_context()
+        if self.insecure:
+            ctx.check_hostname = False
+            ctx.verify_mode = ssl.CERT_NONE
+        return ctx
+
+    def _encode_query_params(self, query_params: Optional[Dict[str, Any]]) -> str:
+        if query_params is None or len(query_params) == 0:
+            return ''
+        encoded_params = urllib.parse.urlencode(query_params)
+        return '?' + encoded_params
 
 
 def _handle_request(func: Callable, name: str, ret_text_on_ok: bool = False) -> Tuple[str, bool]:
     """
     Wraps the error, logging and exception handling.
     """
-    try:
-        resp = func()
-        if resp.status_code != 200:
-            msg = f"{name} failed: {resp.status_code}: {resp.text}"
-            log.error(msg)
-            return msg, False
-        if ret_text_on_ok:
-            return resp.text, True
-        else:
-            return 'OK', True
-
-    except RequestException as exc:
-        msg = f"{name} connection error: {exc}"
+    text, ok = func()
+    if not ok:
+        msg = f"{name} failed: {text}"
         log.error(msg)
         return msg, False
+    if ret_text_on_ok:
+        return text, True
+    else:
+        return 'OK', True
